import logging
import os
import select
import sys
import threading

import _thread
import termios
import tty

from . import responses

logger = logging.getLogger(__name__)


# Initially taken from https://github.com/magmax/python-readchar
def readchar(wait_for_char=0.1):
    old_settings = termios.tcgetattr(sys.stdin)
    tty.setcbreak(sys.stdin.fileno())
    res = b''
    try:
        if select.select([sys.stdin, ], [], [], wait_for_char)[0]:
            res = os.read(sys.stdin.fileno(), 1)
        while select.select([sys.stdin, ], [], [], 0.0)[0]:
            res += os.read(sys.stdin.fileno(), 1)
        if res:
            return res
    finally:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)
    return None


def single_char_with_timeout(timeout=5):
    timer = threading.Timer(timeout, _thread.interrupt_main)
    response = None
    try:
        timer.start()
        while response is None:
            response = readchar()
    except KeyboardInterrupt:
        pass
    timer.cancel()
    return response


def format_track(track):
    return '%s by %s' % (
        track.name,
        ' & '.join(
            artist.name for artist in track.artists
            if artist.name
        )
    )


def format_album(album):
    return '%s by %s [%s]' % (
        album.album.name,
        album.artist.name,
        album.album.year
    )


def sorted_menu_items(items):
    global_items = []
    for key, value in sorted(items):
        if value.destination in responses.ALL:
            global_items.append((key, value))
        else:
            yield key, value
    for key, value in global_items:
        yield key, value

if __name__ == '__main__':
    if sys.argv[-1] == 'wrapper':
        print(single_char_with_timeout())
    else:
        char = readchar(10)
        print(char)
        print(char.decode('utf-8'))



from spotify import TrackAvailability
from collections import namedtuple

from unittest.mock import Mock

Artist = namedtuple('Artist', ('name', ))


class Track(object):
    def __init__(self, name, artists, available=True):
        self.artists = [Artist(artist) for artist in artists]
        self.name = name
        if available:
            self.availability = TrackAvailability.AVAILABLE
        else:
            self.availability = TrackAvailability.UNAVAILABLE


class Playlist(object):
    def __init__(self, name, tracks):
        self.link = Mock()
        self.link.as_playlist.return_value = self
        self.name = name
        self.tracks = tracks



import logging
from collections import defaultdict
import random
import threading
import time
import _thread

import spotify

from .responses import NOOP, UP
from .util import single_char_with_timeout, format_track
from .menus import SavePlaylist

logger = logging.getLogger(__name__)


class Player(object):

    REPEAT_OPTIONS = ['all', 'one']

    shuffle = False
    repeat = REPEAT_OPTIONS[0]
    original_playlist_name = None

    # Initialization and external helpers
    def __init__(self, navigator):
        '''
        Initialize the player. Navigator must be an instance of
        `spoppy.navigation.Leifur`.
        '''
        self.navigator = navigator
        self._initialized = False
        self.end_of_track = None

        self.clear()
        self.actions = {
            b'n': self.next_song,
            b'\xc3\xa6': self.next_song,
            b'p': self.previous_song,
            b'j': self.previous_song,
            b' ': self.play_pause,
            b'u': UP,
            b'q': self.stop_and_clear,
            b'd': self.debug,
            b's': self.toggle_shuffle,
            b'r': self.toggle_repeat,
            b'k': self.backward_10s,
            b'l': self.forward_10s,
            b'h': self.get_help,
            b'x': self.remove_current_song,
            b'w': self.save_as_playlist,
            b'\x1b[A': self.move_song_up,
            b'\x1b[B': self.move_song_down,
        }
        key_names = {
            b' ': 'space',
            b'\x1b[A': 'up arrow',
            b'\x1b[B': 'down arrow',
        }
        self.reversed_actions = defaultdict(list)
        for key, value in self.actions.items():
            if isinstance(value, str):
                action_name = value
            else:
                action_name = value.__name__
            self.reversed_actions[action_name].append(
                key_names.get(key) or key.decode('utf-8')
            )

    def clear(self):
        '''
        Resets all variables used for playlist/queue handling to their default
        values.
        :returns: None
        '''
        self.current_track_idx = 0
        self.current_track = None
        self.seconds_played = 0
        self.play_timestamp = None
        self.song_order = []
        self.current_track_duration = ''
        self.repeat = self.REPEAT_OPTIONS[0]
        self.show_help = False
        self.playlist = None
        self.song_list = []
        self._trigger_redraw = False

    def has_been_loaded(self):
        '''
        Used to determine if some songs are loaded in the player
        :returns: True if there are any songs in the player's queue
        '''
        return bool(len(self.song_list))

    def initialize(self):
        '''
        Initializes some class variables for simpler access to the spotify
        player. `PySpotify` session must have been initialized before calling
        initialize, and initialize must be called before playing anything.
        :returns: None
        '''
        if not self._initialized:
            # For quicker access
            self.session = self.navigator.session
            self.player = self.session.player
            self._initialized = True

    def is_playing(self):
        '''
        Used to determine if the `PySpotify` player is is currently playing
        :returns: True if the player is currently playing
        '''
        return self.player.state == 'playing'

    # UI specific
    def get_duration_from_s(self, s):
        '''
        Formats seconds as "%M:%S"
        :param s: Seconds in int/float
        :returns: s formatted as "%M:%S"
        '''
        # Max length is 59 minutes, 59 seconds
        MAX_LENGTH = 59 * 60 + 59
        if not isinstance(s, (int, float)):
            raise TypeError('Seconds must be int/float')
        elif s < 0:
            raise TypeError('Seconds must be positive')
        elif s > MAX_LENGTH:
            s = MAX_LENGTH
        return '%s:%s' % (
            str(int(s / 60)).zfill(2),
            str(int(s % 60)).zfill(2)
        )

    def get_help_ui(self):
        '''
        Gets menu items explaining the use of hotkeys within the player
        :returns: List of hotkeys and their corresponding actions
        '''
        res = []
        res.append('')
        for action, hotkeys in sorted(self.reversed_actions.items()):
            if self.shuffle and action.startswith('move_song'):
                # Moving songs is not available when we are shuffling
                continue
            res.append(
                '[%s]: %s' % (
                    '/'.join(hotkeys),
                    action
                )
            )
        return res

    def get_progress(self):
        '''
        Get the progress of the currently playing song
        :returns: 4-item tuple, (player_state,
                                 minutes_played,
                                 percent_played,
                                 current_track_duration)
        '''
        seconds_played = self.seconds_played
        if self.play_timestamp is not None:
            # We are actually playing and we have to calculate the number of
            # seconds since we last pressed play
            seconds_played += time.time() - self.play_timestamp

        # pyspotify's duration is in ms
        percent_played = (seconds_played * 1000) / self.current_track.duration
        mins_played = self.get_duration_from_s(seconds_played)

        return (
            self.player.state,
            mins_played,
            percent_played,
            self.current_track_duration
        )

    def get_response(self):
        '''
        Get destination from user and dispatch the event.
        Possibly the event will be handled within the player itself and until
        there comes an event not handled by the player it will block.
        :returns: Destination for the user
        '''
        # This is actually our game loop... because fuck you that's why
        response = NOOP
        while response == NOOP:
            self.session.process_events()
            end_of_track_response = self.check_end_of_track()
            if end_of_track_response:
                return end_of_track_response
            if self.current_track:
                self.navigator.update_progress(*self.get_progress())
            char = single_char_with_timeout(timeout=1.5)
            if char:
                logger.debug('Got some char: %s' % char)
            response = self.actions.get(char, NOOP)
            if response != NOOP:
                if callable(response):
                    evaluated_response = response()
                    if evaluated_response:
                        return evaluated_response
                    # We have handled the response ourselves
                    response = NOOP
                else:
                    return response
            if self._trigger_redraw:
                self._trigger_redraw = False
                return NOOP

    def get_ui(self):
        '''
        Get the UI representing the player's current state
        :returns: List of lines that should be shown. If an item is a 2-item
                  tuple the first one should be left aligned and the second
                  one should be right aligned.
        '''
        res = []
        res.append('Press h for help')
        if self.playlist:
            res.append('Playing playlist: %s' % self.playlist.name)
        if self.show_help:
            res += self.get_help_ui()

        res.append('')

        if self.current_track:
            # We can show number of items - current items - currently playing
            max_number_of_items = self.navigator.get_ui_height() - len(res) - 3
            if self.current_track_idx == 0:
                max_number_of_items += 1
            if self.current_track_idx == len(self.song_list) - 1:
                max_number_of_items += 1
            previous_items_to_show = min([
                int(max_number_of_items / 2),
                self.current_track_idx
            ])
            next_items_to_show = min([
                max_number_of_items - previous_items_to_show,
                len(self.song_list[self.current_track_idx:])
            ])
            total_number_to_show = previous_items_to_show + next_items_to_show
            if total_number_to_show < max_number_of_items:
                previous_items_to_show = min([
                    max_number_of_items - next_items_to_show,
                    len(self.song_list[:self.current_track_idx])

                ])
            right_side_items = [
                'Shuffle on' if self.shuffle else '',
                'Repeat: %s' % self.repeat,
            ]
            songs_to_show = (
                list(range(
                    self.current_track_idx - previous_items_to_show,
                    self.current_track_idx
                )) +
                [self.current_track_idx] +
                list(range(
                    self.current_track_idx + 1,
                    self.current_track_idx + next_items_to_show
                ))
            )
            for song_idx in songs_to_show:
                song = self.get_track_by_idx(song_idx)
                right_side = right_side_items and right_side_items.pop()
                if song_idx == self.current_track_idx:
                    if song_idx != songs_to_show[0]:
                        # Small spacing around current...
                        res.append(('', right_side or ''))
                        right_side = (
                            right_side_items and right_side_items.pop()
                        )
                    formatted_song = '>>>%s' % format_track(song)
                else:
                    formatted_song = format_track(song)
                res.append((formatted_song, right_side or ''))
                if song_idx == self.current_track_idx:
                    if song_idx != songs_to_show[-1]:
                        # Small spacing around current...
                        right_side = (
                            right_side_items and right_side_items.pop()
                        )
                        res.append(('', right_side or ''))
            while right_side_items:
                # This can happend f.x. when we have one song...
                res.append(('', right_side_items.pop()))
        else:
            res.append('No songs found in playlist!')

        return res

    def trigger_redraw(self):
        '''
        Tell the player to trigger a full redraw in the next loop.
        :returns: None
        '''
        self._trigger_redraw = True

    # Event handlers
    def backward_10s(self):
        '''
        Seeks the current song 10 seconds back
        :returns: None
        '''
        if self.play_timestamp is not None:
            self.seconds_played += time.time() - self.play_timestamp
            self.play_timestamp = time.time()
        self.seconds_played -= 10
        if self.seconds_played < 0:
            self.seconds_played = 0
        self.player.seek(int(self.seconds_played * 1000))

    def debug(self):
        '''
        Start a debugger to inspect the player's current state
        :returns: None
        '''
        import pdb
        pdb.set_trace()

    def forward_10s(self):
        '''
        Seeks the current song 10 seconds forward
        :returns: None
        '''
        if self.play_timestamp is not None:
            self.seconds_played += time.time() - self.play_timestamp
            self.play_timestamp = time.time()
        self.seconds_played += 10
        self.player.seek(int(self.seconds_played * 1000))

    def get_help(self):
        '''
        Tell the player to display the help section
        :returns: responses.NOOP
        '''
        self.show_help = not self.show_help
        return NOOP

    def move_song_down(self):
        '''
        Move the currently playing song up one place in the song order.
        If the player is in shuffle mode it does absolutely nothing!
        :returns: responses.NOOP
        '''
        if not self.shuffle:
            i = self.current_track_idx
            k = i + 1
            sl = self.song_list
            if k >= len(sl):
                k = 0
            sl[i], sl[k] = sl[k], sl[i]
            self.current_track_idx = k
        return NOOP

    def move_song_up(self):
        '''
        Move the currently playing song up one place in the song order.
        If the player is in shuffle mode it does absolutely nothing!
        :returns: responses.NOOP
        '''
        if not self.shuffle:
            i = self.current_track_idx
            k = i - 1
            sl = self.song_list
            if k < 0:
                k = len(sl) - 1
            sl[i], sl[k] = sl[k], sl[i]
            self.current_track_idx = k
        return NOOP

    def next_song(self):
        '''
        Plays the next song in the song list
        :returns: responses.NOOP
        '''
        self.current_track_idx = self.get_next_idx()
        self.play_current_song()
        return NOOP

    def play_pause(self):
        '''
        Pauses the current song if it's currently playing, otherwise pauses it.
        :returns: responses.NOOP
        '''
        if not self.is_playing():
            self.player.play()
            self.play_timestamp = time.time()
        else:
            self.player.pause()
            self.seconds_played += time.time() - self.play_timestamp
            self.play_timestamp = None
        return NOOP

    def previous_song(self):
        '''
        Plays the previous song in the song list
        :returns: responses.NOOP
        '''
        self.current_track_idx = self.get_prev_idx()
        self.play_current_song()
        return NOOP

    def remove_current_song(self):
        '''
        Removes the current song from the queue. Note that the song is not
        removed from the playlist itself.
        :returns: responses.NOOP
        '''
        if self.current_track_idx < len(self.song_order):
            idx_in_song_list = self.song_order[self.current_track_idx]
            del self.song_order[self.current_track_idx]

            del self.song_list[idx_in_song_list]

            for idx, item in enumerate(self.song_order):
                if item > idx_in_song_list:
                    self.song_order[idx] -= 1

            if self.current_track_idx >= len(self.song_order):
                self.current_track_idx = 0

            self.play_current_song()
            self.playlist = None
        return NOOP

    def save_as_playlist(self):
        '''
        Prompts the user for a name for a new playlist and allows him to
        save the queue as a new playlist.
        :returns: responses NOOP if the queue has not been modified.
                  menus.SavePlaylist otherwise.
        '''
        def playlist_saved_callback(playlist):
            self.playlist = playlist
            self.original_playlist_name = playlist.name

        res = SavePlaylist(self.navigator)
        res.song_list = self.song_list
        res.original_playlist_name = self.original_playlist_name
        res.callback = playlist_saved_callback
        return res

    def stop_and_clear(self):
        '''
        Stops the current song and clears the current song list, then exits
        the player.
        :returns: responses.UP
        '''
        self.player.unload()
        self.clear()
        return UP

    def toggle_shuffle(self):
        '''
        Puts shuffle mode on/off
        :returns: responses.NOOP
        '''
        # We also have to update the current_track_idx too here since
        # the order is changing
        self.shuffle = not self.shuffle
        currently_playing = None
        if self.current_track_idx < len(self.song_order):
            currently_playing = self.song_order[self.current_track_idx]
        self.set_song_order_by_shuffle()
        if currently_playing in self.song_order:
            self.current_track_idx = self.song_order.index(currently_playing)
        return NOOP

    def toggle_repeat(self):
        '''
        Toggles between available repeat options. See `Player.REPEAT_OPTIONS`
        :returns: responses.NOOP
        '''
        new_idx = self.REPEAT_OPTIONS.index(self.repeat) + 1
        if new_idx >= len(self.REPEAT_OPTIONS):
            new_idx = 0
        self.repeat = self.REPEAT_OPTIONS[new_idx]
        return NOOP

    # Song handling
    def add_to_queue(self, item):
        '''
        Adds item to the end of the current song list. Item can be either
        a single track or a playlist.
        :returns: None
        '''
        if isinstance(item, spotify.Track):
            # Add the newest track_idx to the song order
            self.song_order.append(len(self.song_order))
            # Add the song to the current song list
            self.song_list.append(item)
        elif hasattr(item, 'tracks'):
            for track in item.tracks:
                if track.availability != spotify.TrackAvailability.UNAVAILABLE:
                    self.add_to_queue(track)
        self.playlist = None

    def check_end_of_track(self):
        '''
        Checks if the current song has finished playing and starts playing
        the next song according to the current repeat setting.
        :returns: None
        '''
        if self.end_of_track and self.end_of_track.is_set():
            if self.repeat == 'all':
                self.next_song()
                return NOOP
            elif self.repeat == 'one':
                self.play_current_song()

    def get_next_idx(self):
        '''
        Get the id of the next song. If currently on the last song it returns
        the first one.
        :returns: The id of the next song in queue.
        '''
        if not self.song_order:
            raise RuntimeError('No songs currently in queue')
        current_track_idx = self.current_track_idx + 1
        if current_track_idx >= len(self.song_order):
            current_track_idx = 0
        return current_track_idx

    def get_prev_idx(self):
        '''
        Get the id of the previous song. If currently on the first song it
        returns the last one.
        :returns: The id of the previous song in queue.
        '''
        if not self.song_order:
            raise RuntimeError('No songs currently in queue')
        current_track_idx = self.current_track_idx - 1
        if current_track_idx < 0:
            current_track_idx = len(self.song_order) - 1
        return current_track_idx

    def get_track_by_idx(self, idx):
        '''
        Get the track for the current idx. Uses the shuffle setting to
        determine the song.
        :param idx: The wanted track's position in the queue
        :returns: The `spotify.Track` that is number `idx` in the queue. If
                  `idx` is larger than the number of songs in the queue it
                  returns None
        '''
        try:
            song_index = self.song_order[idx]
            return self.song_list[song_index]
        except IndexError:
            return None

    def load_playlist(self, playlist, shuffle=None):
        '''
        Clears the current song list and replaces it with the playlists
        tracks
        :param playlist: A `spotify.Playlist` to load
        :param shuffle: Shuffle can be explicitly defined. Defaults to using
                        the shuffle setting that was set when the playlist
                        was loaded.
        :returns: None
        '''
        self.clear()
        self.song_list = [
            track for track in
            playlist.tracks
            if track.availability != spotify.TrackAvailability.UNAVAILABLE
        ]
        self.playlist = playlist
        self.original_playlist_name = self.playlist.name
        if shuffle is not None:
            self.shuffle = shuffle
        self.set_song_order_by_shuffle()

    def on_end_of_track(self, session=None):
        '''
        Sets the end of track event and signals the player something has
        happened.
        :param session: A `spotify.Session` (optional, not used)
        :returns: None
        '''
        self.end_of_track.set()
        _thread.interrupt_main()

    def play_current_song(self):
        '''
        Plays the current song
        :returns: None
        '''
        self.player.unload()

        current_track = self.get_track_by_idx(self.current_track_idx)
        if not current_track:
            self.current_track = None
            return

        self.end_of_track = threading.Event()
        self.current_track = current_track.load()

        self.current_track_duration = self.get_duration_from_s(
            self.current_track.duration / 1000
        )

        self.player.load(self.current_track)
        self.play_pause()

        self.seconds_played = 0

        logger.debug('Playing track %s' % self.current_track.name)

        # Register event listeners
        self.session.on(
            spotify.SessionEvent.END_OF_TRACK,
            self.on_end_of_track
        )

    def play_track(self, track_idx):
        '''
        Plays the track that's number `track_idx` in the song list (note, not
        the song order)
        :param track_idx: The position of the desired track to play
        :returns: None
        '''
        self.current_track_idx = self.song_order.index(track_idx)
        self.play_current_song()

    def set_song_order_by_shuffle(self):
        '''
        Based on the current shuffle setting, shuffles the song list or not.
        :returns: None
        '''
        self.song_order = list(range(len(self.song_list)))
        if self.shuffle:
            random.shuffle(self.song_order)



# NOTE: As seen here in https://github.com/mopidy/pyspotify/issues/183
# libspotify's search is broken. Until it either gets fixed (improbable)
# this will have to wrap it up. The goal is to mimick pyspotify's
# search class as best as I can.

import logging
import threading

import requests
from spotify.track import Track, TrackAvailability
from spotify.album import Album

logger = logging.getLogger(__name__)


def search(*args, **kwargs):
    return Search(*args, **kwargs)


class SearchResults(object):
    def __init__(self, term, results, offset, total,
                 previous_page=None, next_page=None):
        self.term = term
        self.results = results
        self.total = total
        self.offset = offset
        self.previous_page = previous_page
        self.next_page = next_page

    def __iter__(self):
        return iter(self.results)

    def __getitem__(self, key):
        return self.results[key]


class Search(threading.Thread):
    ENDPOINTS = {
        # Each entry is a tuple, (HTTP_ENDPOINT, CLS)
        'tracks': (
            '/v1/search?query={query}&offset=0&limit=20&type=track',
            Track
        ),
        'albums': (
            '/v1/search?query={query}&offset=0&limit=20&type=album',
            Album
        ),
        'artists': (
            '/v1/search?query={query}&offset=0&limit=20&type=artist',
            None
        ),
    }
    BASE_URL = 'https://api.spotify.com'

    def __init__(self, session, query='', callback=None,
                 track_offset=0, track_count=20,
                 album_offset=0, album_count=20,
                 artist_offset=0, artist_count=20,
                 playlist_offset=0, playlist_count=20,
                 search_type=None,
                 sp_search=None, add_ref=True, direct_endpoint=None):
        self.session = session
        self.query = query
        self.search_type = search_type
        self.loaded_event = threading.Event()

        endpoint, self.item_cls = self.ENDPOINTS[self.search_type]

        self.endpoint = direct_endpoint or (
            self.BASE_URL + endpoint.format(query=self.query)
        )

        super(Search, self).__init__()

        self.start()

    def run(self):
        try:
            logger.debug('Getting %s' % self.endpoint)
            r = requests.get(self.endpoint)
        except requests.exceptions.ConnectionError:
            self.results = SearchResults(self.query, [], 0, 0)
        else:
            r.raise_for_status()

            response_data = r.json()[self.search_type]

            item_results = self.manipulate_items([
                self.item_cls(self.session, item['uri'])
                for item in response_data['items']
            ])

            self.results = SearchResults(
                self.query,
                item_results,
                response_data['offset'],
                response_data['total'],
                response_data['previous'],
                response_data['next']
            )

        self.loaded_event.set()

    def manipulate_items(self, items):
        if self.search_type == 'albums':
            # Not my fault....
            # See: https://github.com/mopidy/pyspotify/issues/119
            return [
                item.browse().load() for item in items
            ]
        elif self.search_type == 'tracks':
            return [
                item.load() for item in items
                if item.availability != TrackAvailability.UNAVAILABLE
            ]
        raise TypeError('Unknown search type %s' % self.search_type)



import logging
from collections import namedtuple

from spotify import TrackAvailability
from .search import search
from . import responses
from .util import (format_album, format_track, single_char_with_timeout,
                   sorted_menu_items)

logger = logging.getLogger(__name__)

MenuValue = namedtuple('MenuValue', ('name', 'destination'))
MockPlaylist = namedtuple('Playlist', ('name', 'tracks'))


class Options(dict):
    def __init__(self, *args, **kwargs):
        super(Options, self).__init__(*args, **kwargs)
        self._cached_matches = {}
        self._stripped_keys_mapper = {
            key.replace(' ', ''): key
            for key in self
        }
        self.check_unique_keys()

    def __setitem__(self, key, value):
        super(Options, self).__setitem__(key, value)
        self._stripped_keys_mapper[key.replace(' ', '')] = key
        self.check_unique_keys()

    def check_unique_keys(self):
        if not len(self) == len(self._stripped_keys_mapper):
            raise TypeError('Two keys cannot be the same')

    def get_possibilities_from_cache(self, pattern):
        return self._cached_matches.get(pattern)

    def get_possibilities(self, pattern):
        cached_match = self.get_possibilities_from_cache(pattern)
        if cached_match:
            logger.debug('Pattern %s found in cache' % pattern)
        else:
            logger.debug('Trying to match %s' % pattern)
            possibilities_key = []
            possibilities_name = []
            pattern = pattern.lower()
            for key, (name, destination) in self.items():
                if key.lstrip(' ').startswith(pattern):
                    possibilities_key.append(key)
                if (
                    # Only match start of words
                    name.lower().startswith(pattern) or
                    ' %s' % pattern in name.lower()
                ):
                    possibilities_name.append(key)
            logger.debug('possibilities_key: %s' % possibilities_key)
            logger.debug('possibilities_name: %s' % possibilities_name)
            cached_match = self._cached_matches[pattern] = (
                list(set(possibilities_key + possibilities_name))
            )
        return cached_match

    def filter(self, pattern):
        possibilities = self.get_possibilities(pattern)

        return Options({
            key: value
            for key, value in
            self.items()
            if key in possibilities
        })

    def match_best_or_none(self, pattern):
        logger.debug('Trying to match (%s)' % pattern)
        possibilities = self.get_possibilities(pattern)
        logger.debug('Available possibilities: %s' % possibilities)
        if len(possibilities) == 1:
            logger.debug('Exactly one possibility, returning that!')
            return self[possibilities[0]]
        if pattern in self._stripped_keys_mapper:
            logger.debug('Pattern matches stripped key, returning key %s' % (
                self._stripped_keys_mapper[pattern]
            ))
            return self[self._stripped_keys_mapper[pattern]]


class Menu(object):
    INCLUDE_UP_ITEM = True

    BACKSPACE = b'\x7f'
    UP_ARROW = b'\x1b[A'
    DOWN_ARROW = b'\x1b[B'
    PAGE = 0

    def __init__(self, navigator):
        self.navigator = navigator

    def get_options(self):
        raise NotImplementedError('Subclass must define this method')

    def initialize(self):
        self._options = Options(self.get_options())
        self._options['q'] = MenuValue('quit', responses.QUIT)
        if self.INCLUDE_UP_ITEM:
            self._options['u'] = MenuValue('..', responses.UP)
        if self.navigator.player.has_been_loaded():
            self._options['p'] = MenuValue('player', responses.PLAYER)
        self.filter = ''

    def get_response(self):
        response = None
        while response is None:
            response = single_char_with_timeout(60)
            self.navigator.player.check_end_of_track()
        if response == Menu.BACKSPACE:
            self.filter = self.filter[:-1]
            return responses.NOOP
        elif response == Menu.UP_ARROW:
            logger.debug('Got UP_ARROW')
            self.PAGE = max([self.PAGE - 1, 0])
            return responses.NOOP
        elif response == Menu.DOWN_ARROW:
            logger.debug('Got DOWN_ARROW')
            self.PAGE += 1
            return responses.NOOP

        self.filter += response.decode('utf-8')
        if self.filter.endswith('\n'):
            # The user wants to go someplace...
            self.filter = self.filter.replace('\n', '')
            # Gets set as the item to navigate to if we only found one
            is_valid = self.is_valid_response()
            if is_valid:
                # Ok, return
                return is_valid.destination
        # Trigger redraw!
        return responses.NOOP

    def is_valid_response(self):
        return self._options.match_best_or_none(self.filter)

    def get_ui(self):
        if self.filter:
            items = sorted_menu_items(
                self._options.filter(self.filter).items()
            )
        else:
            items = sorted_menu_items(
                self._options.items()
            )
        if not items:
            menu_items = ('No matches for "%s"' % self.filter, )
        else:
            menu_items = tuple(
                self.get_menu_item(key, value.name) for key, value in
                items
            )
            if self.filter:
                is_valid = self.is_valid_response()
                if is_valid:
                    menu_items += (
                        '',
                        'Press [return] to go to (%s)' % is_valid.name
                    )
        if menu_items:
            number_of_menu_items = self.navigator.get_ui_height() - 4
            if len(menu_items) >= number_of_menu_items:
                paginated_menu_items = []
                while not paginated_menu_items:
                    start_idx = self.PAGE * number_of_menu_items
                    end_idx = (self.PAGE + 1) * number_of_menu_items
                    paginated_menu_items = menu_items[start_idx:end_idx]
                    if not paginated_menu_items:
                        self.PAGE -= 1
                menu_items = paginated_menu_items
            else:
                self.PAGE = 0

        above_menu_items = self.get_header()
        return (
            (above_menu_items, '') +
            menu_items +
            ('', 'Query: %s' % self.filter, )
        )

    def get_menu_item(self, key, value):
        return '[%s]: %s' % (key, value)

    def get_header(self):
        return ''


class MainMenu(Menu):
    INCLUDE_UP_ITEM = False

    def get_options(self):
        return {
            'vp': MenuValue(
                'View playlists', PlayListOverview(self.navigator)
            ),
            'st': MenuValue('Search for tracks', TrackSearch(self.navigator)),
            'sa': MenuValue('Search for albums', AlbumSearch(self.navigator)),
        }


class PlayListOverview(Menu):

    def get_options(self):
        def include_playlist(playlist):
            return (
                playlist.name and
                hasattr(playlist, 'link') and
                any([
                    track for track in playlist.tracks
                    if track.availability != TrackAvailability.UNAVAILABLE
                ])
            )
        results = {}
        playlists = self.navigator.session.playlist_container
        playlists = enumerate(
            sorted(
                (
                    playlist for playlist in playlists
                    if include_playlist(playlist)
                ),
                key=lambda x: x.name
            )
        )
        for i, playlist in playlists:
            menu_item = PlayListSelected(self.navigator)
            menu_item.playlist = playlist.link.as_playlist()
            results[str(i+1).rjust(4)] = MenuValue(
                menu_item.playlist.name, menu_item
            )
        return results

    def get_header(self):
        return 'Select a playlist'


class TrackSearchResults(Menu):
    search = None
    paginating = False
    _cached_search_results = []

    def set_initial_results(self, search):
        self.search = search
        self.update_cache()

    def update_cache(self):
        self._cached_search_results.append(self.search)

    def get_cache(self):
        return self._cached_search_results

    def get_response(self):
        if self.paginating:
            self.search.loaded_event.wait()
            self.paginating = False
            return self
        return super(TrackSearchResults, self).get_response()

    def go_to(self, up_down):
        if up_down > 0:
            destination = 'next_page'
        else:
            destination = 'previous_page'

        def inner():
            self.paginating = True

            new_cache_idx = self.get_cache().index(self.search) + up_down

            try:
                self.search = self.get_cache()[new_cache_idx]
                logger.debug('Got search from cache, yahoo!')
            except IndexError:
                direct_endpoint = getattr(
                    self.search.results, destination
                )
                logger.debug('Initiating new search')
                self.search = search(
                    self.navigator.session, self.search.query,
                    search_type=self.search.search_type,
                    direct_endpoint=direct_endpoint
                )
                self.update_cache()
            return self
        return inner

    def select_song(self, track_idx):
        def song_selected():
            track = self.search.results[track_idx]
            if self.navigator.player.is_playing():
                # If the user is currently playing we want him to have the
                # choice to add this song to the list of current songs or
                # start playing this playlist from this song
                menu = SongSelectedWhilePlaying(self.navigator)
                menu.track = track
                return menu
            else:
                self.navigator.player.clear()
                self.navigator.player.add_to_queue(track)
                self.navigator.player.play_track(0)
                return self.navigator.player
        return song_selected

    def get_header(self):
        return 'Search results for %s (total %d results)' % (
            self.search.results.term, self.search.results.total
        )

    def get_res_idx(self, i):
        return i + 1 + self.search.results.offset

    def get_ui(self):
        if self.paginating:
            return 'Loading...'
        return super(TrackSearchResults, self).get_ui()

    def get_options(self):
        if self.paginating:
            return {}
        results = self.get_options_from_search()
        if self.search.results.previous_page:
            results['p'] = MenuValue(
                'Previous', self.go_to(-1)
            )
        if self.search.results.next_page:
            results['n'] = MenuValue(
                'Next', self.go_to(1)
            )
        return results

    def get_options_from_search(self):
        results = {}
        for i, track in enumerate(
            track for track in
            self.search.results
            if track.availability != TrackAvailability.UNAVAILABLE
        ):
            results[str(self.get_res_idx(i)).rjust(4)] = MenuValue(
                format_track(track), self.select_song(i)
            )
        return results


class AlbumSearchResults(TrackSearchResults):
    search = None

    def select_album(self, track_idx):
        def album_selected():
            res = AlbumSelected(self.navigator)
            res.album = self.search.results[track_idx]
            return res
        return album_selected

    def get_options_from_search(self):
        results = {}
        for i, album in enumerate(
            album for album in
            self.search.results
        ):
            results[str(self.get_res_idx(i)).rjust(4)] = MenuValue(
                format_album(album), self.select_album(i)
            )
        return results


class TrackSearch(Menu):
    is_searching = False
    search_pattern = ''
    search = None
    search_type = 'tracks'
    result_cls = TrackSearchResults

    def get_options(self):
        return {}

    def get_search_results(self):
        self.search_pattern = self.filter
        self.search = search(
            self.navigator.session, self.search_pattern,
            search_type=self.search_type
        )
        self.is_searching = True
        return self

    def get_response(self):
        if self.is_searching:
            self.search.loaded_event.wait()
            self.is_searching = False

            search_results = self.result_cls(self.navigator)
            search_results.set_initial_results(self.search)

            self.is_searching = False
            self.search_pattern = ''
            self.search = None

            return search_results
        return super(TrackSearch, self).get_response()

    def is_valid_response(self):
        return super(TrackSearch, self).is_valid_response() or MenuValue(
            None, self.get_search_results
        )

    def get_ui(self):
        if self.is_searching:
            return 'Searching for %s' % self.search_pattern
        else:
            return '\n'.join((
                'Search query: %s' % self.filter,
                '',
                'Press [return] to search',
                '(Pro tip: you can also input "u" to go up or "q" to quit)'
            ))


class AlbumSearch(TrackSearch):
    search_type = 'albums'
    result_cls = AlbumSearchResults


class PlayListSelected(Menu):
    playlist = None
    deleting = False

    def shuffle_play(self):
        self.navigator.player.load_playlist(
            self.playlist,
            shuffle=True
        )
        self.navigator.player.play_current_song()
        return self.navigator.player

    def select_song(self, track_idx):
        def song_selected():
            if self.navigator.player.is_playing():
                # If the user is currently playing we want him to have the
                # choice to add this song to the list of current songs or
                # start playing this playlist from this song
                menu = SongSelectedWhilePlaying(self.navigator)
                menu.playlist = self.playlist
                menu.track = self.get_tracks()[track_idx]
                return menu
            else:
                self.navigator.player.load_playlist(
                    self.playlist
                )
                self.navigator.player.play_track(track_idx)
                return self.navigator.player
        return song_selected

    def add_to_queue(self):
        self.navigator.player.add_to_queue(self.playlist)
        return self.navigator.player

    def delete_playlist(self):
        self.deleting = True
        return self

    def do_delete_playlist(self):
        p_idx = self.navigator.session.playlist_container.index(
            self.playlist
        )
        self.navigator.session.playlist_container.remove_playlist(p_idx)
        return responses.UP

    def cancel_delete_playlist(self):
        self.deleting = False
        return self

    def get_tracks(self):
        return self.playlist.tracks

    def get_name(self):
        return self.playlist.name

    def get_options(self):
        results = {}
        if self.deleting:
            results['y'] = MenuValue('Yes', self.do_delete_playlist)
            results['n'] = MenuValue('No', self.cancel_delete_playlist)
        else:
            for i, track in enumerate(
                track for track in
                self.get_tracks()
                if track.availability != TrackAvailability.UNAVAILABLE
            ):
                results[str(i+1).rjust(4)] = MenuValue(
                    format_track(track), self.select_song(i)
                )
            if results:
                results['sp'] = MenuValue('Shuffle play', self.shuffle_play)
                if self.navigator.player.is_playing():
                    results['add_to_queue'] = MenuValue(
                        'Add %s to queue' % self.get_name(),
                        self.add_to_queue
                    )
            else:
                logger.debug('There are no songs in this playlist!')
            results['x'] = MenuValue('Delete playlist', self.delete_playlist)

        return results

    def get_header(self):
        if self.deleting:
            return 'Are you sure you want to delete playlist [%s]' % (
                self.get_name()
            )
        return 'Playlist [%s] selected' % self.get_name()


class AlbumSelected(PlayListSelected):
    album = None
    _tracks = None

    def initialize(self):
        super(AlbumSelected, self).initialize()
        self.playlist = MockPlaylist(self.get_name(), self.get_tracks())

    def get_tracks(self):
        if not self._tracks:
            self._tracks = self.album.tracks
        return self._tracks

    def get_name(self):
        return format_album(self.album)

    def get_header(self):
        return 'Album [%s] selected' % self.get_name()


class SongSelectedWhilePlaying(Menu):
    playlist = None
    track = None

    def add_to_queue(self):
        self.navigator.player.add_to_queue(self.track)
        return responses.UP

    def replace_current(self):
        self.navigator.player.load_playlist(
            self.playlist
        )
        self.navigator.player.play_track(
            self.playlist.tracks.index(self.track)
        )
        return self.navigator.player

    # Only thing we know here is that the player is currently playing something
    def get_options(self):
        results = {}
        if self.playlist:
            results['replace'] = MenuValue(
                'Replace currently playing with %s' % (self.playlist.name),
                self.replace_current
            )
        results['add_to_queue'] = MenuValue(
            'Add %s to queue' % format_track(self.track),
            self.add_to_queue
        )
        return results

    def get_header(self):
        if self.playlist:
            return 'Song [%s] from playlist [%s] selected' % (
                format_track(self.track), self.playlist.name
            )
        else:
            return 'Song [%s] selected'


class SavePlaylist(Menu):
    song_list = []
    is_saving = False
    callback = None
    original_playlist_name = None

    def get_options(self):
        return {}

    def save_playlist(self):
        self.new_playlist_name = (
            self.filter.strip() or self.original_playlist_name
        )
        self.is_saving = True
        return self

    def get_response(self):
        if self.is_saving:
            edited = False
            if self.new_playlist_name == self.original_playlist_name:
                try:
                    playlist = [
                        playlist for playlist in
                        self.navigator.session.playlist_container
                        if playlist.name == self.new_playlist_name
                    ][0]
                    playlist.load()
                except IndexError:
                    pass
                else:
                    # We found our playlist, now we have to edit it
                    # We kind of want to maintain when the songs were added,
                    # so we have to shuffle songs that were already in the
                    # playlist around
                    # These cases can come up:
                    #   1. Song is in the same place in song list and playlist
                    #   2. Song appears in playlist later than in song list
                    #   3. Song does not appear at all in playlist
                    for song_list_idx, song in enumerate(self.song_list):
                        try:
                            playlist_tracks_idx = playlist.tracks.index(song)
                        except ValueError:
                            # Case 3
                            playlist.add_tracks(song, song_list_idx)
                        else:
                            if playlist_tracks_idx == song_list_idx:
                                # Case 1
                                continue
                            elif playlist_tracks_idx >= song_list_idx:
                                # Case 2
                                playlist.reorder_tracks(
                                    playlist_tracks_idx, song_list_idx
                                )
                    # Now we have to remove all the tracks that are at indexes
                    # larger then the length of the song list
                    playlist.remove_tracks(
                        range(len(self.song_list), len(playlist.tracks))
                    )
                    edited = True
            if not edited:
                playlist = (
                    self.navigator.session.playlist_container.add_new_playlist(
                        self.new_playlist_name
                    )
                )
                playlist.add_tracks(self.song_list)

            playlist.load()
            self.is_saving = False
            while playlist.has_pending_changes:
                self.navigator.session.process_events()
            if self.callback:
                self.callback(playlist)
            return responses.UP
        return super(SavePlaylist, self).get_response()

    def is_valid_response(self):
        return super(SavePlaylist, self).is_valid_response() or (
            (
                self.filter.strip() or self.original_playlist_name
            ) and
            MenuValue(
                None, self.save_playlist
            )
        )

    def get_ui(self):
        if self.is_saving:
            return 'Saving playlist as %s' % self.new_playlist_name
        else:
            return '\n'.join((
                '%d songs to be added to new playlist' % len(self.song_list),
                'The original playlist name was [%s]. Leave the name empty '
                'to replace [%s] with the current song list' % (
                    self.original_playlist_name, self.original_playlist_name
                ) if self.original_playlist_name else '',
                'Playlist name: %s' % self.filter,
                '',
                'Press [return] to save your playlist',
                '(Pro tip: you can also input "u" to go up or "q" to quit)'
            ))



import logging
import os

from appdirs import user_cache_dir

cache_dir = user_cache_dir(appname='spoppy')

LOG_FILE_NAME = os.path.join(
    cache_dir, 'spoppy.log'
)

if not os.path.isdir(cache_dir):
    os.makedirs(cache_dir)

logger = logging.getLogger('spoppy')
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler(LOG_FILE_NAME)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.debug('Logger set up')



import logging

import click

from . import get_version, menus, responses
from .lifecycle import LifeCycle
from .players import Player
from .terminal import get_terminal_size
from .config import clear_config

logger = logging.getLogger(__name__)


class Leifur(object):
    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.player = Player(self)
        self.lifecycle = LifeCycle(username, password, self.player)
        self.session = None
        self.navigating = True
        logger.debug('Leifur initialized')

    def start(self):
        if self.lifecycle.check_pyspotify_logged_in():
            logger.debug('All tokens are a-OK')
            self.session = self.lifecycle.get_pyspotify_client()
            logger.debug('Starting LifeCycle services')
            self.lifecycle.start_lifecycle_services()
            logger.debug('LifeCycle services started')
            self.player.initialize()
            main_menu = menus.MainMenu(self)
            self.navigate_to(main_menu)
        else:
            click.echo(
                'Could not log you in, please check your username (%s) '
                'and password are correct' % self.username
            )
            clear_config()
            logger.debug('Something went wrong, not logged in...')

    def shutdown(self):
        self.lifecycle.shutdown()
        logger.debug('Navigation shutdown complete')

    def navigate_to(self, going):
        logger.debug('navigating to: %s' % going)
        self.session.process_events()
        going.initialize()
        while self.navigating:
            logger.debug('clearing screen...')
            click.clear()
            self.print_header()
            self.print_menu(going.get_ui())
            response = going.get_response()
            logger.debug('Got response %s' % response)
            if callable(response):
                response = response()
                logger.debug('Got response %s after evaluation' % response)
            if response == responses.QUIT:
                click.clear()
                click.echo('Thanks, bye!')
                self.navigating = False
                return
            elif response == responses.UP:
                break
            elif response == responses.NOOP:
                continue
            elif response == responses.PLAYER:
                self.navigate_to(self.player)
            elif response != going:
                self.navigate_to(response)
            # This happens when the `going` instance gets control again. We
            # don't want to remember the query and we want to rebuild the
            # menu's options
            # (and possibly something else?)
            going.initialize()

    def print_header(self):
        click.echo('Spoppy v. %s' % get_version())
        click.echo('Hi there %s' % self.username)
        click.echo('')

    def print_menu(self, menu):
        if isinstance(menu, str):
            click.echo(menu)
        elif isinstance(menu, (list, tuple)):
            for item in menu:
                if isinstance(item, (list, tuple)):
                    if len(item) == 2:
                        click.echo(
                            ''.join((
                                item[0],
                                ' ' * (
                                    self.get_ui_width() -
                                    len(item[0]) -
                                    len(item[1])
                                ),
                                item[1]
                            ))
                        )
                    else:
                        click.echo(item[0])
                else:
                    click.echo(item)
            click.echo('')

    def update_progress(self, status, start, perc, end):
        s = '\r[%s] %s[%s]%s' % (
            status,
            start,
            '%s',
            end or ''
        )
        progress_width = self.get_ui_width() - len(s) + 2
        s = s % ('#'*int(perc*progress_width)).ljust(progress_width)

        click.echo(s, nl=False)

    def get_ui_width(self):
        return get_terminal_size().width

    def get_ui_height(self):
        # We want to return the height allocated to the menu
        return get_terminal_size().height - 4



import getpass
import os

from appdirs import user_cache_dir

CONFIG_FILE_NAME = os.path.join(
    user_cache_dir(appname='spoppy'), '.creds'
)


def get_config():
    if os.path.exists(CONFIG_FILE_NAME):
        with open(CONFIG_FILE_NAME, 'r') as f:
            return [
                line.strip() for line in f.readlines()
            ][:2]
    return None, None


def set_config(username, password):
    with open(CONFIG_FILE_NAME, 'w') as f:
        f.write(username)
        f.write('\n')
        f.write(password)


def get_config_from_user():
    username, password = (
        input('Username: '),
        getpass.getpass('Password: ')
    )
    set_config(username, password)
    return username, password


def clear_config():
    os.remove(CONFIG_FILE_NAME)



import logging
import os
import threading

import spotify
from appdirs import user_cache_dir

from .dbus_listener import DBusListener
from .terminal import ResizeChecker

logger = logging.getLogger(__name__)


class LifeCycle(object):

    user_cache_dir = user_cache_dir(appname='spoppy')

    def __init__(self, username, password, player):
        if not os.path.isdir(self.user_cache_dir):
            os.makedirs(self.user_cache_dir)
        self.player = player
        self.username = username
        self.password = password
        self._pyspotify_session = None
        self._pyspotify_session_loop = None
        self.service_stop_event = threading.Event()
        self.services = [
            DBusListener(self, self.service_stop_event),
            ResizeChecker(self, self.service_stop_event)
        ]

    def start_lifecycle_services(self):
        for service in self.services:
            if service.should_run:
                service.start()
                logger.debug('%s started!' % service)
            else:
                logger.debug('Not running %s' % service)

    def shutdown(self):
        if self._pyspotify_session:
            logger.debug('Logging user out after quit...')
            self._pyspotify_session.logout()
        logger.debug('Closing dbus_listener')
        self.service_stop_event.set()
        while self.services:
            logger.debug('Joining %s' % self.services[0])
            if self.services[0].is_alive():
                # Give it half a second to die
                self.services[0].join(0.5)
            if not self.services[0].is_alive():
                del self.services[0]
        logger.debug('All services joined')
        if self._pyspotify_session_loop:
            self._pyspotify_session_loop.stop()
        logger.debug('Pyspotify session loop stopped')

    def get_pyspotify_client(self):
        return self._pyspotify_session

    def check_pyspotify_logged_in(self):
        logger.debug('Checking if pyspotify is logged in...')
        config = spotify.Config()
        config.user_agent = 'Spoppy'
        config.cache_location = os.path.join(self.user_cache_dir, 'cache')
        config.settings_location = os.path.join(self.user_cache_dir, 'cache')
        config.load_application_key_file(
            os.path.join(os.path.dirname(__file__), 'spotify_appkey.key')
        )
        self._pyspotify_session = spotify.Session(config)
        self._pyspotify_session_loop = spotify.EventLoop(
            self._pyspotify_session
        )
        self._pyspotify_session_loop.start()

        # Connect an audio sink
        spotify.AlsaSink(self._pyspotify_session)

        # Events for coordination
        logged_in = threading.Event()
        # end_of_track = threading.Event()

        def on_connection_state_updated(session):
            if session.connection.state is spotify.ConnectionState.LOGGED_IN:
                logged_in.set()

        # Register event listeners
        self._pyspotify_session.on(
            spotify.SessionEvent.CONNECTION_STATE_UPDATED,
            on_connection_state_updated
        )

        logger.debug('Actually logging in now...')
        self._pyspotify_session.login(self.username, self.password)

        logged_in.wait(6)
        if logged_in.is_set():
            logger.debug('PySpotify logged in!')
            return True
        else:
            logger.warning('PySpotify login failed!')
            return False



import threading
import shutil
from collections import namedtuple
import logging

logger = logging.getLogger(__name__)

TerminalSize = namedtuple('TerminalSize', ('width', 'height'))


def get_terminal_size():
    size = shutil.get_terminal_size((120, 40))
    return TerminalSize(size.columns, size.lines)


class ResizeChecker(threading.Thread):
    CHECK_INTERVAL = .5

    def __init__(self, lifecycle, stop_event, *args):
        self.lifecycle = lifecycle
        self.stop_event = stop_event
        self.last_size = get_terminal_size()
        self.should_run = True
        super(ResizeChecker, self).__init__()

    def run(self):
        logger.debug('ResizeChecker started')
        while not self.stop_event.wait(self.CHECK_INTERVAL):
            new_size = get_terminal_size()
            if self.last_size != get_terminal_size():
                logger.debug('Terminal size changed to %s' % (new_size, ))
                self.lifecycle.player.trigger_redraw()
                self.last_size = new_size



import logging

try:
    import click
    from lockfile import LockFile, LockTimeout
except ImportError:
    click = None

logger = logging.getLogger('spoppy.main')


def get_version():
    return '1.2.0'

if click:
    @click.command()
    @click.argument('username', required=False)
    @click.argument('password', required=False)
    def main(username, password):
        # Ignore error, logging set up in logging utils
        from . import logging_utils
        from .navigation import Leifur
        from .config import get_config, set_config, get_config_from_user

        lock = LockFile('/tmp/spoppy.lock')
        try:
            # Try for 5s to acquire the lock
            lock.acquire(5)
        except LockTimeout:
            click.echo('Could not acquire lock, is spoppy running?')
        else:

            if username and password:
                set_config(username, password)
            else:
                username, password = get_config()
            if not (username and password):
                username, password = get_config_from_user()
            try:
                navigator = Leifur(username, password)
                navigator.start()
            finally:
                navigator.shutdown()
                logger.debug('Finally, bye!')
        finally:
            if lock.i_am_locking():
                lock.release()
else:
    def main(*args, **kwargs):
        print('Something went horribly wrong, missing requirements...')



#!/usr/bin/env python

from __future__ import print_function

import sys

from setuptools import setup
from spoppy import get_version

if sys.version_info < (3, 3):
    print('You need at least python 3.3')
    sys.exit(1)


with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='spoppy',
    version=get_version(),
    description='A lightweight spotify CLI',
    author='Sindri Guðmundsson',
    author_email='sindrigudmundsson@gmail.com',
    url='https://github.com/sindrig/spoppy',
    packages=['spoppy'],
    test_suite='nose.collector',
    install_requires=required,
    include_package_data=True,
    zip_safe=False,
    scripts=[
        'scripts/spoppy'
    ]
)



import unittest
import uuid
from unittest.mock import Mock, patch

from spoppy import menus, responses

from . import utils


class TestOptions(unittest.TestCase):

    def setUp(self):
        self.dct = {
            '1': menus.MenuValue('A', Mock()),
            '2': menus.MenuValue('B', Mock()),
            '3': menus.MenuValue('C', Mock()),
            's': menus.MenuValue('Search', Mock()),
            'kk': menus.MenuValue('þþ', Mock()),
            'ko': menus.MenuValue('þ', Mock()),
            'q': menus.MenuValue('o', Mock()),
            'o': menus.MenuValue('q', Mock()),
        }
        self.op = menus.Options(self.dct)

    def test_options_filter_empty(self):
        self.assertEqual(self.op, self.op.filter(''))

    def test_filter_by_key(self):
        tc = self.op.filter('1')
        self.assertEqual(len(tc), 1)
        self.assertIn('1', tc)
        tc = self.op.filter('k')
        self.assertEqual(len(tc), 2)
        self.assertIn('kk', tc)
        self.assertIn('ko', tc)
        tc = self.op.filter('s')
        self.assertEqual(len(tc), 1)
        self.assertIn('s', tc)

    def test_filter_by_name(self):
        tc = self.op.filter('þ')
        self.assertEqual(len(tc), 2)
        self.assertIn('kk', tc)
        self.assertIn('ko', tc)
        tc = self.op.filter('þþ')
        self.assertEqual(len(tc), 1)
        self.assertIn('kk', tc)
        tc = self.op.filter('Sea')
        self.assertEqual(len(tc), 1)
        self.assertIn('s', tc)

    def test_filter_is_case_insensitive(self):
        self.assertEqual(self.op.filter('Search'), self.op.filter('search'))

    def test_filter_returns_empty_if_no_match(self):
        self.assertEqual(len(self.op.filter('asdf')), 0)

    def test_get_possibilities_from_name_and_key(self):
        tc = self.op.get_possibilities('q')
        self.assertEqual(len(tc), 2)
        self.assertEqual(sorted(tc), sorted(['q', 'o']))

    def test_possibility_not_duplicated(self):
        tc = self.op.get_possibilities('s')
        self.assertEqual(len(tc), 1)
        self.assertIn('s', tc)

    def test_possiblities_with_spaces(self):
        op = menus.Options({
            '    a': menus.MenuValue('þ', Mock()),
            'b    ': menus.MenuValue('þ', Mock()),
            '    c    ': menus.MenuValue('þ', Mock()),
        })
        for key in 'a', 'b', 'c':
            tc = op.get_possibilities(key)
            self.assertEqual(len(tc), 1)
            self.assertNotEqual(tc, [key])
            self.assertIn(key, tc[0])

    def test_possibilities_only_start_of_words(self):
        # Also testing for case insesitivity
        op = menus.Options({
            'match1': menus.MenuValue('Sindri', Mock()),
            'not1': menus.MenuValue('iSindri', Mock()),
            'match2': menus.MenuValue('This is sindri', Mock()),
            'not2': menus.MenuValue('This is notSindri', Mock()),
        })
        possibilities = op.get_possibilities('Sindri')
        self.assertEqual(sorted(possibilities), sorted(['match1', 'match2']))

    def test_matches_by_correct_key(self):
        op = menus.Options({
            'k': menus.MenuValue('1', Mock()),
            'kk': menus.MenuValue('2', Mock()),
            'kkk': menus.MenuValue('3', Mock()),
        })
        best = op.match_best_or_none('k')
        self.assertEqual(best.name, '1')
        best = op.match_best_or_none('kk')
        self.assertEqual(best.name, '2')
        best = op.match_best_or_none('kkk')
        self.assertEqual(best.name, '3')

    def test_matches_by_correct_padded_key(self):
        op = menus.Options({
            '    a': menus.MenuValue('1', Mock()),
            'b    ': menus.MenuValue('2', Mock()),
            '    c    ': menus.MenuValue('3', Mock()),
            ' s i ': menus.MenuValue('4', Mock()),
        })
        best = op.match_best_or_none('a')
        self.assertEqual(best.name, '1')
        best = op.match_best_or_none('b')
        self.assertEqual(best.name, '2')
        best = op.match_best_or_none('c')
        self.assertEqual(best.name, '3')
        best = op.match_best_or_none('si')
        self.assertEqual(best.name, '4')

    def test_check_unique_keys(self):
        with self.assertRaises(TypeError):
            menus.Options({
                'a': menus.MenuValue('þ', Mock()),
                ' a': menus.MenuValue('k', Mock())
            })
        with self.assertRaises(TypeError):
            self.op['   1'] = menus.MenuValue('1', Mock())


class MenuTests(unittest.TestCase):

    def setUp(self):
        self.navigator = Mock()
        self.navigator.get_ui_height.return_value = 100

        class SubMenu(menus.Menu):
            def get_options(self):
                return {}

        self.submenu = SubMenu(self.navigator)

    def test_must_be_subclassed(self):
        m = menus.Menu(self.navigator)
        with self.assertRaises(NotImplementedError):
            m.get_options()

    def test_global_options_correct(self):
        self.submenu.INCLUDE_UP_ITEM = False
        self.navigator.player.has_been_loaded.return_value = False
        self.submenu.initialize()
        included_items = [
            value.destination for value in self.submenu._options.values()
        ]
        self.assertEqual(len(included_items), 1)
        self.assertIn(responses.QUIT, included_items)
        self.assertNotIn(responses.UP, included_items)
        self.assertNotIn(responses.PLAYER, included_items)

        self.submenu.INCLUDE_UP_ITEM = True
        self.navigator.player.has_been_loaded.return_value = False
        self.submenu.initialize()
        included_items = [
            value.destination for value in self.submenu._options.values()
        ]
        self.assertEqual(len(included_items), 2)
        self.assertIn(responses.QUIT, included_items)
        self.assertIn(responses.UP, included_items)
        self.assertNotIn(responses.PLAYER, included_items)

        self.submenu.INCLUDE_UP_ITEM = False
        self.navigator.player.has_been_loaded.return_value = True
        self.submenu.initialize()
        included_items = [
            value.destination for value in self.submenu._options.values()
        ]
        self.assertEqual(len(included_items), 2)
        self.assertIn(responses.QUIT, included_items)
        self.assertNotIn(responses.UP, included_items)
        self.assertIn(responses.PLAYER, included_items)

        self.submenu.INCLUDE_UP_ITEM = True
        self.navigator.player.has_been_loaded.return_value = True
        self.submenu.initialize()
        included_items = [
            value.destination for value in self.submenu._options.values()
        ]
        self.assertEqual(len(included_items), 3)
        self.assertIn(responses.QUIT, included_items)
        self.assertIn(responses.UP, included_items)
        self.assertIn(responses.PLAYER, included_items)

    def test_filter_initialized_correctly(self):
        self.assertFalse(hasattr(self.submenu, 'filter'))
        self.submenu.initialize()
        self.assertTrue(hasattr(self.submenu, 'filter'))
        self.assertEqual(self.submenu.filter, '')

    @patch('spoppy.menus.single_char_with_timeout')
    def test_pagination_keys(self, patched_chargetter):
        self.assertEqual(self.submenu.PAGE, 0)

        patched_chargetter.return_value = menus.Menu.DOWN_ARROW
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.PAGE, 1)

        patched_chargetter.return_value = menus.Menu.UP_ARROW
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.PAGE, 0)

        patched_chargetter.return_value = menus.Menu.UP_ARROW
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.PAGE, 0)

    @patch('spoppy.menus.single_char_with_timeout')
    def test_backspace(self, patched_chargetter):
        self.submenu.initialize()

        patched_chargetter.return_value = b'a'
        self.assertEqual(self.submenu.filter, '')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'a')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'aa')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'aaa')

        patched_chargetter.return_value = menus.Menu.BACKSPACE
        self.assertEqual(self.submenu.filter, 'aaa')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'aa')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'a')
        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, '')

    @patch('spoppy.menus.Menu.is_valid_response')
    @patch('spoppy.menus.single_char_with_timeout')
    def test_return(self, patched_chargetter, patched_is_valid):
        destination = 'DESTINATION'
        patched_is_valid.return_value = menus.MenuValue('TEST', destination)
        self.submenu.initialize()

        patched_chargetter.return_value = b'\n'
        self.assertEqual(self.submenu.get_response(), destination)
        patched_is_valid.assert_called_once_with()

    @patch('spoppy.menus.single_char_with_timeout')
    def test_checks_for_end_of_track(self, patched_chargetter):
        patched_chargetter.side_effect = [None, None, b'a']

        self.submenu.initialize()

        self.assertEqual(self.submenu.get_response(), responses.NOOP)
        self.assertEqual(self.submenu.filter, 'a')
        self.assertEqual(
            self.navigator.player.check_end_of_track.call_count, 3
        )

    @patch('spoppy.menus.Options.match_best_or_none')
    def test_is_valid_uses_options(self, patched_match_best_or_none):
        patched_match_best_or_none.return_value = 'RETVAL'
        self.submenu.initialize()
        self.submenu.filter = 'ASDF'
        self.assertEqual(self.submenu.is_valid_response(), 'RETVAL')
        patched_match_best_or_none.assert_called_once_with('ASDF')

    @patch('spoppy.menus.Options.filter')
    def test_ui_filters_items(self, patched_filter):
        self.submenu.initialize()
        patched_filter.return_value = self.submenu._options
        self.submenu.get_ui()
        patched_filter.assert_not_called()
        self.submenu.filter = 'a'
        self.submenu.get_ui()
        patched_filter.assert_called_once_with('a')

    @patch('spoppy.menus.sorted_menu_items')
    def test_no_matches_warning_shown(self, patched_sorter):
        self.submenu.initialize()
        self.submenu.filter = ''
        patched_sorter.return_value = []
        ui = self.submenu.get_ui()
        has_filter_in_line = [line for line in ui if 'No matches' in line]
        self.assertEqual(len(has_filter_in_line), 1)

    @patch('spoppy.menus.Menu.get_menu_item')
    def test_uses_get_menu_item(self, patched_get_menu_item):
        self.submenu.initialize()
        self.submenu.filter = ''
        patched_get_menu_item.return_value = 'OHAI'

        ui = self.submenu.get_ui()

        self.assertEqual(
            patched_get_menu_item.call_count,
            len([line for line in ui if line == 'OHAI'])
        )

    def test_shows_indicator_if_one_match(self):
        self.submenu.filter = 'a'
        self.submenu.get_options = Mock()
        self.submenu.get_options.return_value = menus.Options({
            'the_key': menus.MenuValue('sindri', Mock()),
            'foo': menus.MenuValue('foo', Mock()),
            'bar': menus.MenuValue('bar', Mock()),
        })
        self.submenu.initialize()

        ui = self.submenu.get_ui()
        self.assertEqual(len([line for line in ui if 'sindri' in line]), 1)
        self.submenu.filter = 'the_key'
        ui = self.submenu.get_ui()
        self.assertEqual(len([line for line in ui if 'sindri' in line]), 2)

    def test_pagination_ui(self):
        option_indicator = 'THIS IS AN OPTION'

        random_options = {
            str(uuid.uuid4()): menus.MenuValue(option_indicator, Mock())
            for i in range(1000)
        }
        get_options = Mock()
        get_options.return_value = random_options
        self.submenu.get_options = get_options
        self.submenu.initialize()

        seen_options = 0

        last_page = -1
        while last_page != self.submenu.PAGE:
            ui = self.submenu.get_ui()
            if self.submenu.PAGE == last_page:
                break
            seen_options += len([
                line for line in ui
                if option_indicator in line
            ])

            last_page = self.submenu.PAGE
            self.submenu.PAGE += 1

        self.assertEqual(seen_options, len(random_options))


class TestSubMenus(unittest.TestCase):

    def setUp(self):
        self.navigator = Mock()

    def get_playlist_selected(self):
        ps = menus.PlayListSelected(self.navigator)
        tracks = [
            utils.Track('Lazarus', ['David Bowie']),
            utils.Track('Best song ever', ['Sindri'], False),
            utils.Track('Blackstar', ['David Bowie']),
            utils.Track('Ziggy Stardust', ['David Bowie']),
        ]
        ps.playlist = utils.Playlist('Playlist', tracks)
        return ps

    def test_playlist_overview_shows_all_playlists(self):
        self.playlists = [
            utils.Playlist('A', [utils.Track('foo', ['bar'])]),
            utils.Playlist('B', [utils.Track('foo', ['bar'])]),
            utils.Playlist('C', [utils.Track('foo', ['bar'])]),
        ]

        class Session(object):
            playlist_container = self.playlists
        self.navigator.session = Session()
        pov = menus.PlayListOverview(self.navigator)
        options = menus.Options(pov.get_options())
        self.assertTrue(
            all(
                isinstance(value.destination, menus.PlayListSelected)
                for value in options.values()
            )
        )
        for playlist in self.playlists:
            self.assertIsNotNone(options.match_best_or_none(playlist.name))

    def test_playlist_overview_does_not_show_invalid_playlists(self):
        self.playlists = [
            # This is the only one that should be shown
            utils.Playlist('A', [utils.Track('foo', ['bar'])]),
            utils.Playlist('B', []),
            utils.Playlist(
                'C', [utils.Track('foo', ['bar'], available=False)]
            ),
            utils.Playlist('D', []),
            utils.Playlist('', []),
        ]
        del self.playlists[1].link

        class Session(object):
            playlist_container = self.playlists
        self.navigator.session = Session()
        pov = menus.PlayListOverview(self.navigator)
        options = menus.Options(pov.get_options())
        self.assertEqual(len(options), 1)
        self.assertEqual(
            list(options.values())[0].destination.playlist,
            self.playlists[0]
        )

    def test_playlist_selected_does_not_fail_on_empty_playlist(self):
        ps = menus.PlayListSelected(self.navigator)
        ps.playlist = utils.Playlist('asdf', [])
        # Only delete available
        self.assertEqual(len(ps.get_options()), 1)

    def test_playlist_selected_contains_only_valid_tracks(self):
        ps = self.get_playlist_selected()
        options = menus.Options(ps.get_options())

        self.assertIsNotNone(options.match_best_or_none('1'))
        self.assertIsNotNone(options.match_best_or_none('2'))
        self.assertIsNotNone(options.match_best_or_none('3'))
        self.assertIsNone(options.match_best_or_none('4'))

    def test_shows_shuffle_play(self):
        ps = self.get_playlist_selected()
        options = menus.Options(ps.get_options())

        destinations = [value.destination for value in options.values()]
        self.assertIn(ps.shuffle_play, destinations)

    def test_shows_add_to_queue_if_playing(self):
        ps = self.get_playlist_selected()

        self.navigator.player.is_playing.return_value = False
        options = menus.Options(ps.get_options())

        destinations = [value.destination for value in options.values()]
        self.assertNotIn(ps.add_to_queue, destinations)

        self.navigator.player.is_playing.return_value = True
        options = menus.Options(ps.get_options())

        destinations = [value.destination for value in options.values()]
        self.assertIn(ps.add_to_queue, destinations)

    def test_select_song(self):
        ps = self.get_playlist_selected()
        song_selected = ps.select_song(0)

        self.navigator.player.is_playing.return_value = False
        self.assertEqual(song_selected(), self.navigator.player)
        self.navigator.player.play_track.assert_called_once_with(0)

        self.navigator.player.is_playing.return_value = True
        song_selected_result = song_selected()
        self.assertIsInstance(
            song_selected_result, menus.SongSelectedWhilePlaying
        )
        self.assertEqual(song_selected_result.playlist, ps.playlist)
        self.assertEqual(song_selected_result.track, ps.playlist.tracks[0])
        self.assertEqual(self.navigator.player.play_track.call_count, 1)


class TestSearch(unittest.TestCase):

    def setUp(self):
        self.navigator = Mock()

    @patch('spoppy.menus.Menu.get_response')
    def test_uses_parent_get_response(self, patched_get_response):
        for cls in (
            menus.TrackSearchResults,
            menus.AlbumSearchResults,
            menus.TrackSearch,
            menus.AlbumSearch
        ):
            patched_get_response.reset_mock()
            patched_get_response.return_value = 'foobar'
            menu = cls(self.navigator)
            self.assertEqual(menu.get_response(), 'foobar')
            patched_get_response.assert_called_once_with()

    @patch('spoppy.menus.TrackSearchResults.update_cache')
    def test_updates_cache_on_init(self, patched_update):
        search = 'foobar'
        menu = menus.TrackSearchResults(self.navigator)
        menu.set_initial_results(search)
        patched_update.assert_called_once_with()

    def test_get_update_cache(self):
        search = 'foobar'
        menu = menus.TrackSearchResults(self.navigator)
        self.assertEqual(len(menu.get_cache()), 0)
        menu.search = search
        menu.update_cache()
        self.assertIn(search, menu.get_cache())

    @patch('spoppy.menus.TrackSearchResults.search')
    def test_resets_paginating(self, patched_search):
        patched_search.loaded_event.wait.return_value = True
        menu = menus.TrackSearchResults(self.navigator)
        menu.paginating = True
        self.assertEqual(menu.get_response(), menu)
        self.assertFalse(menu.paginating)
        patched_search.loaded_event.wait.assert_called_once_with()

    @patch('spoppy.menus.TrackSearchResults.update_cache')
    @patch('spoppy.menus.search')
    @patch('spoppy.menus.TrackSearchResults.get_cache')
    def test_go_to_from_cache(
        self, patched_cache,
        patched_search, patched_update
    ):
        patched_cache.return_value = [Mock(), Mock()]

        menu = menus.TrackSearchResults(self.navigator)
        menu.search = patched_cache.return_value[0]
        # next_page
        callback = menu.go_to(1)
        self.assertEqual(callback(), menu)
        self.assertEqual(menu.search, patched_cache.return_value[1])
        self.assertTrue(menu.paginating)
        patched_update.assert_not_called()
        patched_search.assert_not_called()

        # previous_page
        callback = menu.go_to(-1)
        self.assertTrue(callable(callback))
        self.assertEqual(callback(), menu)
        self.assertEqual(menu.search, patched_cache.return_value[0])
        self.assertTrue(menu.paginating)
        patched_update.assert_not_called()
        patched_search.assert_not_called()

    @patch('spoppy.menus.TrackSearchResults.update_cache')
    @patch('spoppy.menus.search')
    @patch('spoppy.menus.TrackSearchResults.get_cache')
    def test_go_to_from_search(
        self, patched_cache,
        patched_search, patched_update
    ):
        patched_cache.return_value = [Mock()]
        patched_search.return_value = Mock()

        menu = menus.TrackSearchResults(self.navigator)
        menu.search = patched_cache.return_value[0]

        callback = menu.go_to(1)
        self.assertTrue(callable(callback))
        self.assertEqual(callback(), menu)
        self.assertEqual(menu.search, patched_search.return_value)
        self.assertTrue(menu.paginating)
        patched_update.assert_called_once_with()
        # Don't check for how it was called, at least not at the moment
        self.assertEqual(patched_search.call_count, 1)

    @patch('spoppy.menus.TrackSearchResults.search')
    def test_select_song_while_playing(self, patched_self_search):
        patched_self_search.results = ['foo']
        self.navigator.player.is_playing.return_value = True

        menu = menus.TrackSearchResults(self.navigator)

        callback = menu.select_song(0)

        self.assertTrue(callable(callback))
        res = callback()
        self.assertIsInstance(res, menus.SongSelectedWhilePlaying)
        self.assertEqual(res.track, 'foo')

    @patch('spoppy.menus.TrackSearchResults.search')
    def test_select_song_while_paused(self, patched_self_search):
        patched_self_search.results = ['foo']
        self.navigator.player.is_playing.return_value = False

        menu = menus.TrackSearchResults(self.navigator)

        callback = menu.select_song(0)

        self.assertTrue(callable(callback))
        res = callback()
        self.assertEqual(res, self.navigator.player)
        self.navigator.player.clear.assert_called_once_with()
        self.navigator.player.add_to_queue.assert_called_once_with('foo')
        self.navigator.player.play_track.assert_called_once_with(0)

    @patch('spoppy.menus.TrackSearchResults.search')
    def test_get_res_idx(self, patched_self_search):
        menu = menus.TrackSearchResults(self.navigator)

        for i in range(0, 5, 20):
            patched_self_search.results.offset = i
            self.assertEqual(menu.get_res_idx(0), i+1)

    @patch('spoppy.menus.Menu.get_ui')
    def test_returns_different_ui_while_paginating(self, patched_get_ui):
        menu = menus.TrackSearchResults(self.navigator)

        first_one = menu.get_ui()
        menu.paginating = True
        second_one = menu.get_ui()
        self.assertNotEqual(first_one, second_one)
        patched_get_ui.assert_called_once_with()

    @patch('spoppy.menus.TrackSearchResults.get_options_from_search')
    def test_get_options(self, patched_get_from_search):
        patched_get_from_search.side_effect = [
            {}, {}, {}, {}
        ]

        menu = menus.TrackSearchResults(self.navigator)
        menu.search = Mock()
        menu.search.results.previous_page = True
        menu.search.results.next_page = True

        menu.paginating = True
        self.assertEqual(len(menu.get_options()), 0)

        menu.paginating = False
        self.assertEqual(len(menu.get_options()), 2)

        menu.search.results.previous_page = False
        self.assertEqual(len(menu.get_options()), 1)

        menu.search.results.next_page = False
        self.assertEqual(len(menu.get_options()), 0)


class TestPlaylistSaver(unittest.TestCase):

    def setUp(self):
        self.navigator = Mock()

    def test_returns_different_ui_while_paginating(self):
        menu = menus.SavePlaylist(self.navigator)
        menu.filter = ''
        menu.song_list = []

        first_one = menu.get_ui()
        menu.is_saving = True
        menu.new_playlist_name = ''
        second_one = menu.get_ui()
        self.assertNotEqual(first_one, second_one)

    @patch('spoppy.menus.Menu.get_response')
    def test_uses_parent_get_response(self, patched_get_response):
        patched_get_response.reset_mock()
        patched_get_response.return_value = 'foobar'
        menu = menus.SavePlaylist(self.navigator)
        self.assertEqual(menu.get_response(), 'foobar')
        patched_get_response.assert_called_once_with()

    def test_saves_playlist(self):
        playlist_mock = Mock()
        (
            self.navigator.session.playlist_container
            .add_new_playlist.return_value
        ) = playlist_mock
        playlist_mock.has_pending_changes = False

        menu = menus.SavePlaylist(self.navigator)
        menu.is_saving = True
        menu.new_playlist_name = 'foobar'
        menu.song_list = [1, 2, 3]
        menu.callback = Mock()

        self.assertEqual(menu.get_response(), responses.UP)

        (
            menu.navigator.session.playlist_container.add_new_playlist
        ).assert_called_once_with(menu.new_playlist_name)

        playlist_mock.add_tracks.assert_called_once_with(menu.song_list)
        playlist_mock.load.assert_called_once_with()

        menu.callback.assert_called_once_with(playlist_mock)



import logging
from urllib.parse import parse_qs, urlparse

from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn

logger = logging.getLogger(__name__)
HTTP_ADDRESS = '0.0.0.0'
HTTP_PORT = 8157


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""


def get_handler(callback_on_token):
    class HTTPHandler(BaseHTTPRequestHandler):

        def do_GET(self):
            """Respond to a GET request."""
            logger.debug('Received GET request %s', self.path)
            parts = parse_qs(urlparse(self.path).query)
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.write("<html><head><title>Spotify oAuth</title></head>")
            self.write("<body><p>This is a token response from Spotify.</p>")
            self.write('<p>%r</p>' % parts)
            self.write("<p>You accessed path: %s</p>" % self.path)
            self.write("</body></html>")
            if parts:
                logger.debug('Got parts %s, calling callback', parts)
                callback_on_token(parts)

        def write(self, line):
            self.wfile.write(line.encode('utf-8'))

    return HTTPHandler


def run(ready_callback):
    def on_ready_callback(parts):
        logger.debug('On ready called with parts %s, shutting down', parts)
        server.shutdown()
        ready_callback(parts)
    server = ThreadedHTTPServer(
        (
            HTTP_ADDRESS,
            HTTP_PORT
        ),
        get_handler(on_ready_callback))
    server.serve_forever()



import logging
import threading

import _thread

from .util import format_track

logger = logging.getLogger(__name__)

try:
    import dbus
    import dbus.service
    import dbus.mainloop.glib
    DBusServiceObject = dbus.service.Object
except ImportError:
    DBusServiceObject = object
    dbus = None
    logger.warning(
        'DBus not installed, you won\'t be able to control the '
        'player via DBus'
    )

try:
    import gobject
except ImportError:
    gobject = None
    logger.warning(
        'gobject not installed, you won\'t be able to control the '
        'player via DBus'
    )

if dbus:
    class SpoppyDBusService(DBusServiceObject):

        def __init__(self, lifecycle):
            self.lifecycle = lifecycle

        def run(self):
            gobject.threads_init()
            dbus.mainloop.glib.threads_init()
            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
            bus_name = dbus.service.BusName(
                "com.spoppy",
                dbus.SessionBus()
            )
            super(SpoppyDBusService, self).__init__(
                bus_name, "/com/spoppy"
            )

            self._loop = gobject.MainLoop()
            self._loop.run()

        def stop(self):
            self._loop.quit()

        @dbus.service.method(
            "com.spoppy",
            in_signature='', out_signature='b'
        )
        def PlayPause(self):
            self.lifecycle.player.play_pause()
            return True

        @dbus.service.method(
            "com.spoppy",
            in_signature='', out_signature='s'
        )
        def Previous(self):
            self.lifecycle.player.previous_song()
            self.lifecycle.player.trigger_redraw()
            _thread.interrupt_main()
            return format_track(self.lifecycle.player.current_track)

        @dbus.service.method(
            "com.spoppy",
            in_signature='', out_signature='s'
        )
        def Next(self):
            self.lifecycle.player.next_song()
            self.lifecycle.player.trigger_redraw()
            _thread.interrupt_main()
            return format_track(self.lifecycle.player.current_track)

        @dbus.service.method(
            "com.spoppy",
            in_signature='', out_signature='s'
        )
        def Current(self):
            if self.lifecycle.player.current_track:
                return format_track(self.lifecycle.player.current_track)
            return ''


class DBusListener(threading.Thread):
    def __init__(self, lifecycle, stop_event, *args):
        self.lifecycle = lifecycle
        self.stop_event = stop_event
        self.should_run = dbus and gobject
        if not self.should_run:
            logger.warning(
                'DBusListener thread aborting because of missing dependencies'
            )
        super(DBusListener, self).__init__()

    def run(self):
        self.service = SpoppyDBusService(self.lifecycle)
        self.service_thread = threading.Thread(
            target=self.service.run
        )
        self.service_thread.start()

        logger.debug('Service started, waiting for kill signal')
        self.stop_event.wait()
        logger.debug('Kill signal received, stopping service')
        self.service.stop()
        while True:
            logger.debug('Joining service thread with timeout 10s')
            self.service_thread.join(10)
            if not self.service_thread.is_alive():
                break



import unittest
from collections import namedtuple
from unittest.mock import MagicMock, Mock, patch

import spotify
from spoppy import players, responses

from . import utils


class TestPlayer(unittest.TestCase):
    def setUp(self):
        self.navigation = Mock()
        self.player = players.Player(self.navigation)
        self.player.initialize()

    def tearDown(self):
        del self.player
        del self.navigation

    def test_has_been_loaded(self):
        self.assertFalse(self.player.has_been_loaded())
        self.player.song_list = [
            utils.Track('', '')
        ]
        self.assertTrue(self.player.has_been_loaded())

    def test_shows_playlist_name(self):
        playlist_name = 'Playlist 1'
        ui = self.player.get_ui()
        self.assertEqual(
            len([line for line in ui if playlist_name in line]),
            0
        )
        self.player.playlist = utils.Playlist(playlist_name, [])
        ui = self.player.get_ui()
        self.assertEqual(
            len([line for line in ui if playlist_name in line]),
            1
        )

    @patch('spoppy.players.Player.get_help_ui')
    def test_shows_help(self, patched_show_help):
        self.player.get_ui()
        self.assertEqual(patched_show_help.call_count, 0)
        self.player.get_help()
        self.player.get_ui()
        self.assertEqual(patched_show_help.call_count, 1)

    def test_shows_all_available_actions_in_help(self):
        help_items = self.player.get_help_ui()
        actions = []
        for item in help_items:
            if '[' in item and ']' in item:
                actions.append(item.split(':')[-1].lstrip(' '))
        for action in self.player.reversed_actions:
            self.assertIn(action, actions)

    def test_get_duration_from_s(self):
        test_cases = {
            0: '00:00',
            9999999999999: '59:59',
            59: '00:59',
            60: '01:00',
            157: '02:37'
        }
        for seconds, expected in test_cases.items():
            self.assertEqual(
                self.player.get_duration_from_s(seconds), expected
            )

    def test_get_duration_from_s_raises(self):
        with self.assertRaises(TypeError):
            self.player.get_duration_from_s(None)
        with self.assertRaises(TypeError):
            self.player.get_duration_from_s(-1)
        with self.assertRaises(TypeError):
            self.player.get_duration_from_s('01:57')

    @patch('spoppy.players.time')
    def test_get_progress_while_playing(self, patched_time):
        self.player.player = Mock()
        self.player.player.state = 'playing'
        # This would amount to 60 seconds played
        patched_time.time.return_value = 30
        self.player.play_timestamp = 0
        self.player.seconds_played = 30
        self.player.current_track = Mock()
        self.player.current_track.duration = 120 * 1000

        state, mins_played, perc_played, duration = self.player.get_progress()
        self.assertEqual(state, self.player.player.state)
        self.assertEqual(mins_played, '01:00')
        self.assertEqual(perc_played, 0.5)

    def test_get_progress_while_paused(self):
        self.player.player = Mock()
        self.player.player.state = 'paused'
        # This would amount to 30 seconds played
        self.player.seconds_played = 30
        self.player.current_track = Mock()
        self.player.current_track.duration = 120 * 1000

        state, mins_played, perc_played, duration = self.player.get_progress()
        self.assertEqual(state, self.player.player.state)
        self.assertEqual(mins_played, '00:30')
        self.assertEqual(perc_played, 0.25)

    @patch('spoppy.players.time')
    def test_seek_backwards(self, patched_time):
        patched_time.time.return_value = 30

        self.player.player = Mock()
        self.player.play_timestamp = 0

        self.assertIsNone(self.player.backward_10s())

        self.assertEqual(self.player.play_timestamp, 30)
        self.assertEqual(self.player.seconds_played, 20)
        self.player.player.seek.assert_called_once_with(20 * 1000)

    def test_seek_backwards_doesnt_seek_negative(self):
        self.player.seconds_played = 1

        self.player.backward_10s()

        self.assertEqual(self.player.seconds_played, 0)
        self.player.player.seek.assert_called_once_with(0)

    @patch('spoppy.players.time')
    def test_seek_forwards(self, patched_time):
        patched_time.time.return_value = 30

        self.player.player = Mock()
        self.player.play_timestamp = 0

        self.assertIsNone(self.player.forward_10s())

        self.assertEqual(self.player.play_timestamp, 30)
        self.assertEqual(self.player.seconds_played, 40)
        self.player.player.seek.assert_called_once_with(40 * 1000)

    def test_seek_doesnt_set_play_timestamp_if_paused(self):
        self.player.play_timestamp = None
        self.player.forward_10s()
        self.assertIsNone(self.player.play_timestamp)

    @patch('spoppy.players.Player.is_playing')
    @patch('spoppy.players.time')
    def test_plays_when_paused(self, patched_time, patched_is_playing):
        self.player.player = Mock()
        patched_is_playing.return_value = False
        patched_time.time.return_value = 100

        self.assertEqual(self.player.play_pause(), responses.NOOP)

        self.player.player.play.assert_called_once_with()
        self.player.player.pause.assert_not_called()

    @patch('spoppy.players.Player.is_playing')
    @patch('spoppy.players.time')
    def test_pauses_when_playing(self, patched_time, patched_is_playing):
        self.player.player = Mock()
        self.player.play_timestamp = 0
        patched_is_playing.return_value = True
        patched_time.time.return_value = 100

        self.assertEqual(self.player.play_pause(), responses.NOOP)

        self.player.player.pause.assert_called_once_with()
        self.player.player.play.assert_not_called()

        self.assertEqual(self.player.seconds_played, 100)
        self.assertIsNone(self.player.play_timestamp)

    @patch('spoppy.players.Player.play_current_song')
    @patch('spoppy.players.Player.get_prev_idx')
    def test_play_prev_song(self, patched_get_prev_idx, patched_play_current):
        patched_get_prev_idx.return_value = 7

        self.assertEqual(self.player.previous_song(), responses.NOOP)

        self.assertEqual(self.player.current_track_idx, 7)
        patched_play_current.assert_called_once_with()

    @patch('spoppy.players.Player.play_current_song')
    @patch('spoppy.players.Player.get_next_idx')
    def test_play_next_song(self, patched_get_next_idx, patched_play_current):
        patched_get_next_idx.return_value = 7

        self.assertEqual(self.player.next_song(), responses.NOOP)

        self.assertEqual(self.player.current_track_idx, 7)
        patched_play_current.assert_called_once_with()

    @patch('spoppy.players.Player.play_current_song')
    def test_remove_current_track(self, patched_play_current):
        track_to_remove = utils.Track('foo', ['bar'])
        song_list = [
            utils.Track('A', ['A']),
            utils.Track('B', ['B']),
            track_to_remove,
            utils.Track('C', ['C']),
            utils.Track('D', ['D']),
        ]
        playlist = utils.Playlist('Playlist 1', song_list)

        self.player.load_playlist(playlist)

        self.assertEqual(len(self.player.song_list), len(song_list))
        self.assertEqual(len(self.player.song_order), len(song_list))
        self.assertIsNotNone(self.player.playlist)

        self.player.current_track_idx = song_list.index(track_to_remove)

        self.assertIn(track_to_remove, self.player.song_list)

        self.assertEqual(self.player.remove_current_song(), responses.NOOP)

        self.assertNotIn(track_to_remove, self.player.song_list)

        self.assertEqual(len(self.player.song_list), len(song_list) - 1)
        self.assertEqual(len(self.player.song_order), len(song_list) - 1)

        patched_play_current.assert_called_once_with()
        self.assertIsNone(self.player.playlist)

    @patch('spoppy.players.Player.play_current_song')
    def test_starts_beginning_if_last_song_removed(self, patched_play_current):
        track_to_remove = utils.Track('foo', ['bar'])
        song_list = [
            utils.Track('A', ['A']),
            utils.Track('B', ['B']),
            utils.Track('C', ['C']),
            utils.Track('D', ['D']),
            track_to_remove,
        ]
        playlist = utils.Playlist('Playlist 1', song_list)

        self.player.load_playlist(playlist)

        self.player.current_track_idx = song_list.index(track_to_remove)

        self.assertEqual(self.player.remove_current_song(), responses.NOOP)

        patched_play_current.assert_called_once_with()
        self.assertEqual(self.player.current_track_idx, 0)

    @patch('spoppy.players.Player.play_current_song')
    def test_remove_song_doesnt_raise_with_empty_q(self, patched_play_current):
        song_list = [
        ]
        playlist = utils.Playlist('Playlist 1', song_list)

        self.player.load_playlist(playlist)
        self.player.current_track_idx = 0
        self.assertEqual(self.player.remove_current_song(), responses.NOOP)

        patched_play_current.assert_not_called()
        self.assertEqual(self.player.current_track_idx, 0)

    def test_shuffle(self):
        # Testing that shuffle maintains the currently playing song
        # is kind of impossible, just testing that the shuffle flag toggles
        self.assertEqual(self.player.shuffle, False)
        self.assertEqual(self.player.toggle_shuffle(), responses.NOOP)
        self.assertEqual(self.player.shuffle, True)
        self.assertEqual(self.player.toggle_shuffle(), responses.NOOP)
        self.assertEqual(self.player.shuffle, False)

    @patch('spoppy.players.Player.clear')
    def test_stop_and_clear(self, patched_clear):
        self.player.player = Mock()
        self.assertEqual(self.player.stop_and_clear(), responses.UP)
        patched_clear.assert_called_once_with()
        self.player.player.unload.assert_called_once_with()

    def test_toggle_repeat(self):
        seen_repeat_flags = []
        for i in range(len(players.Player.REPEAT_OPTIONS)):
            self.assertEqual(self.player.toggle_repeat(), responses.NOOP)
            seen_repeat_flags.append(self.player.repeat)
        self.assertEqual(
            sorted(players.Player.REPEAT_OPTIONS),
            sorted(seen_repeat_flags)
        )

    def test_add_track_to_queue(self):
        track = MagicMock(spec=spotify.Track)
        self.player.playlist = 'foo'
        self.assertIsNone(self.player.add_to_queue(track))
        self.assertIn(track, self.player.song_list)
        self.assertIsNone(self.player.playlist)

    def test_add_playlist_to_queue(self):
        tracks = [
            MagicMock(spec=spotify.Track),
            MagicMock(spec=spotify.Track),
            MagicMock(spec=spotify.Track),
        ]
        for track in tracks:
            track.availability = spotify.TrackAvailability.AVAILABLE
        playlist = MagicMock(spec=spotify.Playlist)
        playlist.tracks = tracks
        self.player.playlist = 'foo'
        self.assertIsNone(self.player.add_to_queue(playlist))
        for track in tracks:
            self.assertIn(track, self.player.song_list)
        self.assertIsNone(self.player.playlist)

    @patch('spoppy.players.Player.next_song')
    @patch('spoppy.players.Player.play_current_song')
    def test_check_end_of_track_doesnt_do_anything_if_song_is_playing(
        self, patched_play_current, patched_next_song
    ):
        self.player.end_of_track = Mock()
        self.player.end_of_track.is_set.return_value = False
        self.player.check_end_of_track()
        patched_play_current.assert_not_called()
        patched_next_song.assert_not_called()

    @patch('spoppy.players.Player.next_song')
    @patch('spoppy.players.Player.play_current_song')
    def test_check_end_of_track_plays_next_song(
        self, patched_play_current, patched_next_song
    ):
        self.player.end_of_track = Mock()
        self.player.end_of_track.is_set.return_value = True
        self.player.repeat = 'all'
        self.player.check_end_of_track()
        patched_play_current.assert_not_called()
        patched_next_song.assert_called_once_with()

    @patch('spoppy.players.Player.next_song')
    @patch('spoppy.players.Player.play_current_song')
    def test_check_end_of_track_plays_current_song(
        self, patched_play_current, patched_next_song
    ):
        self.player.end_of_track = Mock()
        self.player.end_of_track.is_set.return_value = True
        self.player.repeat = 'one'
        self.player.check_end_of_track()
        patched_play_current.assert_called_once_with()
        patched_next_song.assert_not_called()

    def test_get_next_prev_idx_raises_with_empty_queue(self):
        with self.assertRaises(RuntimeError):
            self.player.get_next_idx()
        with self.assertRaises(RuntimeError):
            self.player.get_prev_idx()

    def test_get_next_idx_wraps(self):
        self.player.song_order = [1, 2, 3]
        self.player.current_track_idx = 2
        self.assertEqual(self.player.get_next_idx(), 0)

    def test_get_prev_idx_wraps(self):
        self.player.song_order = [1, 2, 3]
        self.player.current_track_idx = 0
        self.assertEqual(self.player.get_prev_idx(), 2)

    @patch('spoppy.players.Player.set_song_order_by_shuffle')
    def test_load_playlist(self, patched_set_shuffle):
        song_list = [
            utils.Track('A', ['A']),
            utils.Track('B', ['B']),
            utils.Track('C', ['C']),
            utils.Track('D', ['D']),
        ]
        playlist = utils.Playlist('Playlist 1', song_list)

        self.player.load_playlist(playlist)

        self.assertEqual(self.player.playlist, playlist)
        self.assertEqual(len(self.player.song_list), len(song_list))

        for i in range(len(song_list)):
            # Test that order is maintained
            self.assertEqual(song_list[i], self.player.song_list[i])

    def test_load_playlist_sets_shuffle(self):
        self.player.load_playlist(utils.Playlist('foo', []), shuffle=True)
        self.assertEqual(self.player.shuffle, True)
        self.player.load_playlist(utils.Playlist('foo', []))
        self.assertEqual(self.player.shuffle, True)
        self.player.load_playlist(utils.Playlist('foo', []), shuffle=False)
        self.assertEqual(self.player.shuffle, False)
        self.player.load_playlist(utils.Playlist('foo', []))
        self.assertEqual(self.player.shuffle, False)

    def test_load_playlist_does_not_load_unplayable_tracks(self):
        track_a = utils.Track('A', ['A'])
        track_b = utils.Track('C', ['C'])
        song_list = [
            track_a,
            utils.Track('B', ['B'], available=False),
            track_b,
            utils.Track('D', ['D'], available=False),
        ]
        playlist = utils.Playlist('Playlist 1', song_list)

        self.player.load_playlist(playlist)

        self.assertEqual(self.player.playlist, playlist)
        self.assertEqual(len(self.player.song_list), 2)
        self.assertIn(track_a, self.player.song_list)
        self.assertIn(track_b, self.player.song_list)

    @patch('spoppy.players._thread')
    def test_on_end_of_track(self, patched__thread):
        self.player.end_of_track = Mock()
        self.player.on_end_of_track()
        self.player.end_of_track.set.assert_called_once_with()
        patched__thread.interrupt_main.assert_called_once_with()

    @patch('spoppy.players.threading')
    @patch('spoppy.players.Player.get_track_by_idx')
    @patch('spoppy.players.Player.get_duration_from_s')
    @patch('spoppy.players.Player.play_pause')
    def test_play_current_song(
        self, patched_play_pause, patched_get_duration, patched_get_track,
        patched_threading
    ):
        self.player.player = Mock()
        self.player.session = Mock()
        patched_track = Mock()
        TrackLoaded = namedtuple('TrackLoaded', ('duration', 'name'))
        track_loaded = TrackLoaded(1, 'foo')
        patched_track.load.return_value = track_loaded
        patched_threading.Event.return_value = 'Event'
        patched_get_track.return_value = patched_track
        patched_get_duration.return_value = 'Duration'

        self.assertIsNone(self.player.play_current_song())

        # Unloads previously playing song
        self.player.player.unload.assert_called_once_with()

        self.assertEqual(self.player.end_of_track, 'Event')
        patched_track.load.assert_called_once_with()
        self.assertEqual(self.player.current_track, track_loaded)
        self.assertEqual(self.player.current_track_duration, 'Duration')

        patched_play_pause.assert_called_once_with()
        self.assertEqual(self.player.seconds_played, 0)

        self.player.session.on.assert_called_once_with(
            spotify.SessionEvent.END_OF_TRACK,
            self.player.on_end_of_track
        )

    @patch('spoppy.players.threading')
    @patch('spoppy.players.Player.get_track_by_idx')
    def test_play_current_song_handles_empty_queue(
        self, patched_get_track, patched_threading
    ):
        self.player.player = Mock()
        patched_get_track.return_value = None
        self.player.play_current_song()
        self.assertIsNone(self.player.current_track)
        patched_threading.Event.assert_not_called()

    @patch('spoppy.players.random')
    def test_set_song_order_by_shuffle(self, patched_random):
        original = [1, 2, 3, 4, 5]
        self.player.song_list = [1, 2, 3, 4, 5]
        self.player.shuffle = False
        self.player.set_song_order_by_shuffle()

        self.assertEqual(
            len(self.player.song_list), len(self.player.song_order)
        )
        self.assertEqual(
            len(original), len(self.player.song_order)
        )

        patched_random.shuffle.assert_not_called()

        self.player.shuffle = True

        self.player.set_song_order_by_shuffle()

        self.assertEqual(
            len(self.player.song_list), len(self.player.song_order)
        )
        self.assertEqual(
            len(original), len(self.player.song_order)
        )

        patched_random.shuffle.assert_called_once_with(self.player.song_order)

    @patch('spoppy.players.Player.play_current_song')
    def test_play_track_by_idx(self, patched_play_current):
        self.player.song_order = [0, 1, 2, 3]

        self.player.play_track(0)
        patched_play_current.assert_called_once_with()
        self.assertEqual(self.player.current_track_idx, 0)

        patched_play_current.reset_mock()

        self.player.song_order = [2, 1, 3, 0]

        self.player.play_track(0)
        patched_play_current.assert_called_once_with()
        self.assertEqual(self.player.current_track_idx, 3)

        with self.assertRaises(ValueError):
            self.player.play_track(None)

    @patch('spoppy.players.SavePlaylist')
    def test_save_as_playlist(self, patched_saveplaylist):
        SavePlaylist = Mock()
        patched_saveplaylist.return_value = SavePlaylist
        self.player.playlist = 'Something'

        self.assertEqual(self.player.save_as_playlist(), SavePlaylist)
        self.assertEqual(self.player.song_list, SavePlaylist.song_list)
        self.assertTrue(callable(SavePlaylist.callback))

        playlist = Mock()
        playlist.name = 'foobar'
        SavePlaylist.callback(playlist)
        self.assertEqual(self.player.playlist, playlist)
        self.assertEqual(self.player.original_playlist_name, 'foobar')

        self.player.playlist = None

        self.assertEqual(self.player.save_as_playlist(), SavePlaylist)
        self.assertEqual(self.player.song_list, SavePlaylist.song_list)

        self.assertTrue(callable(SavePlaylist.callback))

        SavePlaylist.callback(playlist)
        self.assertEqual(self.player.playlist, playlist)
        self.assertEqual(self.player.original_playlist_name, 'foobar')
